"""
https://en.wikipedia.org/wiki/USB
https://github.com/pyusb/pyusb/blob/master/docs/tutorial.rst
"""
import time
from typing import Dict, List
import usb.core
import usb.util
from usb.legacy import CLASS_HUB

# lsusb -d 2e8a:0005 -v

RP2_VENDOR = 0x2E8A
RP2_PRODUCT_BOOT_MODE = 0x0003
RP2_PRODUCT_APPLICATION_MODE = 0x0005

"""
<DEVICE ID 0bda:5411 on Bus 003 Address 046>
  parent=<DEVICE ID 0bda:5411 on Bus 003 Address 044>
<DEVICE ID 0bda:5411 on Bus 003 Address 049>
  parent=<DEVICE ID 0bda:5411 on Bus 003 Address 045>
<DEVICE ID 0bda:5411 on Bus 003 Address 047>
  parent=<DEVICE ID 0bda:5411 on Bus 003 Address 045>
<DEVICE ID 0bda:5411 on Bus 003 Address 045>
  parent=<DEVICE ID 0bda:5411 on Bus 003 Address 044>
<DEVICE ID 0bda:5411 on Bus 003 Address 044>
  parent=<DEVICE ID 1d6b:0002 on Bus 003 Address 001>

<DEVICE ID 0bda:0411 on Bus 004 Address 010>
  parent=<DEVICE ID 0bda:0411 on Bus 004 Address 008>
<DEVICE ID 0bda:0411 on Bus 004 Address 013>
  parent=<DEVICE ID 0bda:0411 on Bus 004 Address 009>
<DEVICE ID 0bda:0411 on Bus 004 Address 011>
  parent=<DEVICE ID 0bda:0411 on Bus 004 Address 009>
<DEVICE ID 0bda:0411 on Bus 004 Address 009>
  parent=<DEVICE ID 0bda:0411 on Bus 004 Address 008>
<DEVICE ID 0bda:0411 on Bus 004 Address 008>
  parent=<DEVICE ID 1d6b:0003 on Bus 004 Address 001>

Root hubs:
<DEVICE ID 1d6b:0002 on Bus 001 Address 001>
<DEVICE ID 1d6b:0003 on Bus 002 Address 001>
<DEVICE ID 1d6b:0002 on Bus 003 Address 001>
<DEVICE ID 1d6b:0003 on Bus 004 Address 001>
*********** duration=11.724s

"""


class find_class:
    def __init__(self, class_: int):
        self._class = class_

    def __call__(self, device):
        # first, let's check the device
        if device.bDeviceClass == self._class:
            return True
        # ok, transverse all devices to find an
        # interface that matches our class
        for cfg in device:
            # find_descriptor: what's it?
            intf = usb.util.find_descriptor(cfg, bInterfaceClass=self._class)
            if intf is not None:
                return True

        return False


class Universe:
    def __init__(self):
        self.busses: Dict[int, usb.Bus] = {}

    def populate(self, hub_devices: List[usb.Device]) -> bool:
        """
        Try to build the tree from the busses, root_hubs, etc.
        return False if no elements could be added.
        """
        added = False
        for hub_device in hub_devices:
            if hub_device.bus not in self.busses:
                self.busses[hub_device.bus] = Bus(nr=hub_device.bus)
            bus = self.busses[hub_device.bus]

            if bus.populate(hub_device=hub_device):
                added = True
        return added

    def print_tree(self) -> None:
        for bus in sorted(self.busses):
            self.busses[bus].print_tree()


class Bus:
    def __init__(self, nr: int):
        self.nr = nr
        self.root_hub: "Hub" = None
        self.hubs: Dict[str:"Hub"] = {}

    def add_hub_(self, hub: "Hub") -> None:
        assert hub.repr_hub_device not in self.hubs
        self.hubs[repr(hub.hub_device)] = hub

    def populate(self, hub_device: usb.Device) -> bool:
        """
        Try to build the tree from the busses, root_hubs, etc.
        return False if no elements could be added.
        """
        if hub_device.parent is None:
            # This is a root hub
            if self.root_hub is not None:
                # The root hub has already been added
                assert hub_device is self.root_hub.hub_device
                return False
            assert len(self.hubs) == 0
            self.root_hub = Hub(bus=self, hub_device=hub_device)
            return True

        if repr(hub_device.parent) in self.hubs:
            # The parent was already added
            parent = self.hubs[repr(hub_device.parent)]
            # Try to add ourselves
            return parent.populate(hub_device)

        # The parent is not populated yet
        return True

    def print_tree(self) -> None:
        print(f"BUS {self.nr}")
        self.root_hub.print_tree(indent="  ")


class Hub:
    def __init__(self, bus: Bus, hub_device: usb.Device):
        self.bus = bus
        self.hub_device = hub_device
        self.repr_hub_device = repr(hub_device)
        self.children: Dict[str:Hub] = {}
        self.port_child: Dict[int:Hub] = {}

        bus.add_hub_(self)

    def _find_port(self, child_hub: "Hub") -> int:
        pass

    def populate(self, hub_device: usb.Device) -> bool:
        """
        Try to build the tree from the busses, root_hubs, etc.
        return False if no elements could be added.
        """
        assert repr(hub_device.parent) == self.repr_hub_device
        if repr(hub_device) in self.children:
            # We have already be added
            return False

        hub = Hub(bus=self.bus, hub_device=hub_device)
        assert hub.repr_hub_device not in self.children
        self.children[repr(hub_device)] = hub
        if False:
            port = self._find_port(hub)
            assert port not in self.port_child
            self.port_child[port] = hub
        return True

    @property
    def children_sorted(self) -> List["Hub"]:
        return self.children.values()

    def print_tree(self, indent: str) -> None:
        print(
            f"{indent} {self.hub_device!r} port_number={self.hub_device.port_number}({self.hub_device.port_numbers})"
        )
        for child in self.children_sorted:
            child.print_tree(indent + "  ")


def main():
    begin_s = time.monotonic()

    hub_devices = usb.core.find(find_all=1, custom_match=find_class(CLASS_HUB))
    hub_devices = list(hub_devices)

    if True:
        u = Universe()
        while u.populate(hub_devices=hub_devices):
            u.print_tree()
            print("-------------")
    for hub in hub_devices:
        if hub.port_number is None:
            print("Päng")
        else:
            print(f"  port_number={hub.port_number}({hub.port_numbers})")
        print(hub)
        continue
        try:
            serial_number = hub.serial_number
        except ValueError:
            serial_number = None
        print(f"{hub!r}")
        if hub.parent is not None:
            print(f"  parent={hub.parent!r}")
        if False:
            print(
                f"  {hub.idVendor:04X}:{hub.idProduct:04X} serial_number={serial_number} port_number={hub.port_number}({hub.port_numbers})"
            )
            print(
                f"  bus={hub.bus} parent.address={hub.address} bcdDevice={hub.bcdDevice} "
            )
            if hub.parent is not None:
                print(f"  parent.bcdDevice={hub.parent.bcdDevice}")

    print(f"*********** duration={time.monotonic()-begin_s:0.3f}s")


if __name__ == "__main__":
    main()



import logging
import pathlib
import subprocess
import time

DIRECTORY_OF_THIS_FILE = pathlib.Path(__file__).parent

logger = logging.getLogger(__file__)


def subprocess_run(args: list[str], timeout_s: float = 10.0) -> str:
    args_text = " ".join(args)

    begin_s = time.monotonic()
    proc = subprocess.run(
        args,
        check=False,
        capture_output=True,
        text=True,
        cwd=str(DIRECTORY_OF_THIS_FILE),
        timeout=timeout_s,
    )

    logger.info(
        f"EXEC {args_text}: returncode={proc.returncode}, duration={time.monotonic()-begin_s:0.3f}s"
    )
    logger.debug(f"  stdout: {proc.stdout}")
    logger.debug(f"  stderr: {proc.stderr}")

    if proc.returncode != 0:
        logger.warning(f"{args_text}: Failed: returncode={proc.returncode}")
        logger.warning(f"STDERR: {proc.stderr}")
        logger.warning(f"STDOUT: {proc.stdout}")
        raise ValueError(
            f"{args_text}:\n\nstderr: {proc.stderr}\n\nstdout: {proc.stdout}"
        )

    return proc.stdout



from usbhubctl import Topology, get_real_topology

from .util_subprocess import subprocess_run


def lsusb() -> Topology:
    args = ["lsusb", "--tree", "--verbose"]
    lsusb_output = subprocess_run(args=args)
    return get_real_topology(lsusb_output=lsusb_output)



import dataclasses

import pytest

from usbhubctl import Hub, Topology
from usbhubctl.known_hubs import rsh_a10, rsh_a16

_EFFECTIVE_TOPOLOGY_WITH_RSH_A10 = """
05E3:0626 2-3
0BDA:0411 2-3.3
0BDA:0411 2-3.3.3
0BDA:0411 2-3.3.4
05E3:0610 3-6
0BDA:5411 3-6.3
0BDA:5411 3-6.3.3
0BDA:5411 3-6.3.4
"""

_EFFECTIVE_TOPOLOGY_WITH_RSH_A16 = """
05E3:0626 1-3
0BDA:0411 1-3.1
0BDA:0411 1-3.1.3
0BDA:0411 1-3.1.3.3
0BDA:0411 1-3.1.3.4
0BDA:0411 1-3.1.4
05E3:0610 2-6
0BDA:5411 2-6.1
0BDA:5411 2-6.1.3
0BDA:5411 2-6.1.3.3
0BDA:5411 2-6.1.3.4
0BDA:5411 2-6.1.4
"""


@dataclasses.dataclass(frozen=True)
class FindHubTestParam:
    hub: Hub
    is_usb2: bool
    effective_topology: str
    expected_connected_hub: str

    def __post_init__(self) -> None:
        assert isinstance(self.hub, Hub)
        assert isinstance(self.is_usb2, bool)
        assert isinstance(self.effective_topology, str)
        assert isinstance(self.expected_connected_hub, str)

    @property
    def testname(self) -> str:
        return f"{self.hub.model}_{'usb2' if self.is_usb2 else 'usb3'}_{self.expected_connected_hub}"


@pytest.mark.parametrize(
    "param",
    (
        FindHubTestParam(rsh_a10, True, _EFFECTIVE_TOPOLOGY_WITH_RSH_A10, "2-3.3"),
        FindHubTestParam(rsh_a10, False, _EFFECTIVE_TOPOLOGY_WITH_RSH_A10, "3-6.3"),
        FindHubTestParam(rsh_a16, True, _EFFECTIVE_TOPOLOGY_WITH_RSH_A16, "1-3.1"),
        FindHubTestParam(rsh_a16, False, _EFFECTIVE_TOPOLOGY_WITH_RSH_A16, "2-6.1"),
    ),
    ids=lambda param: param.testname,
)
def test_find_hub(param: FindHubTestParam) -> None:
    effective_topology = Topology.short_with_product_factory(param.effective_topology)
    connected_hubs = param.hub.find_connected_hubs(
        is_usb2=param.is_usb2,
        actual_usb_topology=effective_topology,
    )

    assert connected_hubs.short == param.expected_connected_hub


_EXPECTED_SHORT_WITH_PRODUCT_RSH_A10_USB2 = """
0BDA:0411 None-
0BDA:0411 None-3
0BDA:0411 None-4
"""

_EXPECTED_SHORT_WITH_PRODUCT_RSH_A10_USB3 = """
0BDA:5411 None-
0BDA:5411 None-3
0BDA:5411 None-4
"""

_EXPECTED_SHORT_WITH_PRODUCT_RSH_A16_USB2 = """
0BDA:0411 None-
0BDA:0411 None-3
0BDA:0411 None-3.3
0BDA:0411 None-3.4
0BDA:0411 None-4
"""

_EXPECTED_SHORT_WITH_PRODUCT_RSH_A16_USB3 = """
0BDA:5411 None-
0BDA:5411 None-3
0BDA:5411 None-3.3
0BDA:5411 None-3.4
0BDA:5411 None-4
"""


@dataclasses.dataclass(frozen=True)
class TopologyTestParam:
    hub: Hub
    is_usb2: bool
    expected_short_with_product: str

    def __post_init__(self) -> None:
        assert isinstance(self.hub, Hub)
        assert isinstance(self.is_usb2, bool)
        assert isinstance(self.expected_short_with_product, str)

    @property
    def testname(self) -> str:
        return f"{self.hub.model}_{'usb2' if self.is_usb2 else 'usb3'}"


@pytest.mark.parametrize(
    "param",
    (
        TopologyTestParam(rsh_a10, True, _EXPECTED_SHORT_WITH_PRODUCT_RSH_A10_USB2),
        TopologyTestParam(rsh_a10, False, _EXPECTED_SHORT_WITH_PRODUCT_RSH_A10_USB3),
        TopologyTestParam(rsh_a16, True, _EXPECTED_SHORT_WITH_PRODUCT_RSH_A16_USB2),
        TopologyTestParam(rsh_a16, False, _EXPECTED_SHORT_WITH_PRODUCT_RSH_A16_USB3),
    ),
    ids=lambda param: param.testname,
)
def test_topology(
    param: TopologyTestParam,
) -> None:
    short_with_product = param.hub.get_topology(
        is_usb2=param.is_usb2
    ).short_with_product

    assert short_with_product == param.expected_short_with_product.strip()


@pytest.mark.parametrize(
    "plug_number,expected_internal_path",
    (
        (
            1,
            [1],
        ),
        (
            2,
            [2],
        ),
        (16, [3, 3, 4]),
    ),
)
def test_plug_rsh_a16(
    plug_number: int,
    expected_internal_path: list[int],
) -> None:
    plug = rsh_a16.get_plug(plug_number=plug_number)
    assert plug.internal_path == expected_internal_path


if __name__ == "__main__":
    test_find_rsh_a16()



"""
Specify the topology of a Hub
"""

from usbhubctl import DualProductId, Hub, HubChip

_PRODUCT1 = DualProductId.parse("2109:0822-2109:2822")
_PRODUCT2 = DualProductId.parse("0bda:0411-0bda:5411")


rsh_st07c = Hub(
    manufacturer="RSHTECH",
    model="RSH-ST07C",
    comment="Plugs 1,2,3: Power control does NOT work!",
    plug_count=7,
    hub_chip=HubChip(
        _PRODUCT1,
        plug_or_chip=[
            1,
            2,
            3,
            HubChip(_PRODUCT2, plug_or_chip=[4, 5, 6, 7]),
        ],
    ),
)



"""
Test live usb hubs
"""

import pytest

from usbhubctl import DualConnectedHub, DualConnectedHubs
from usbhubctl.known_hubs import rsh_a10, rsh_a16, rsh_a107, rsh_st07c
from usbhubctl.util_logging import init_logging


def get_one_or_skip(connected_hubs: DualConnectedHubs) -> DualConnectedHub:
    try:
        return connected_hubs.get_one()
    except Exception as e:
        pytest.skip(msg=f"Could not connect to '{connected_hubs.hub}': {e}")


@pytest.mark.live
def test_toggle_rsh_a10() -> None:
    connected_hubs = rsh_a10.find_connected_dualhubs()

    connected_hub = get_one_or_skip(connected_hubs)
    for plug in connected_hub.connected_plugs:
        plug.power(on=True)
    for plug in connected_hub.connected_plugs:
        plug.power(on=False)


@pytest.mark.live
def test_toggle_rsh_a16() -> None:
    connected_hubs = rsh_a16.find_connected_dualhubs()

    connected_hub = get_one_or_skip(connected_hubs)
    for plug in connected_hub.connected_plugs:
        plug.power(on=True)

    for plug in connected_hub.connected_plugs:
        plug.power(on=False)


@pytest.mark.live
def test_toggle_rsh_a16_plug1() -> None:
    connected_hubs = rsh_a16.find_connected_dualhubs()

    connected_hub = get_one_or_skip(connected_hubs)
    plug = connected_hub.get_plug(1)
    plug.power(on=True)
    plug.power(on=False)


@pytest.mark.live
def test_toggle_rsh_st07c() -> None:
    connected_hubs = rsh_st07c.find_connected_dualhubs()

    connected_hub = get_one_or_skip(connected_hubs)
    for plug in connected_hub.connected_plugs:
        plug.power(on=True)
    for plug in connected_hub.connected_plugs:
        plug.power(on=False)


@pytest.mark.live
def test_toggle_rsh_a107() -> None:
    connected_hubs = rsh_a107.find_connected_dualhubs()

    connected_hub = get_one_or_skip(connected_hubs)
    for plug in connected_hub.connected_plugs:
        plug.power(on=True)
    for plug in connected_hub.connected_plugs:
        plug.power(on=False)


if __name__ == "__main__":
    init_logging()

    # test_toggle_rsh_a10()
    # test_toggle_rsh_a107()
    test_toggle_rsh_st07c()


"""
/sys/devices/pci0000:00/0000:00:14.0/usb3/3-6/3-6.1/power
echo on > control


/sys/devices/pci0000:00/0000:00:14.0/usb3/3-6/3-6.1/3-6.1:1.0/3-6.1-port1/power

/sys/devices/pci0000:00/0000:00:14.0/usb3/3-6/3-6.1/3-6.1:1.0/3-6.1-port1/disable
  echo 1 / 0
  ==> Only works after 'uhubctl -l 3-6.1 -p 1 --action on'

uhubctl -l 3-6.1 -p 1 --action on


------------------------
cd /sys/bus/usb/devices/3-6.1:1.0
cd 3-6.1-port1

=========usb2

cd /sys/bus/usb/devices/2-3.1:1.0
cd 2-3.1-port1
echo 1 > disable

=====================
echo 1 > /sys/bus/usb/devices/2-3.1:1.0/2-3.1-port1/disable
echo 1 > /sys/bus/usb/devices/3-6.1:1.0/3-6.1-port1/disable

"""



"""
Elaborate the real topology using `lsusb -tv
"""

import itertools
import re
from collections.abc import Iterator

from usbhubctl import Path, ProductId, Topology

RE_BUS1 = re.compile(
    r"^/:  Bus (?P<bus>\d+).Port (?P<port>\d+): Dev \d+, Class=root_hub,"
)
"""
Example: '/:  Bus 002.Port 001: Dev 001, Class=root_hub, Driver=xhci_hcd/4p, 10000M'
"""

RE_PORT1 = re.compile(
    r"^(?P<indent1> +\|__ )Port (?P<port>\d+).*?Class=(?P<class>.*?),"
)
"""
Example: '    |__ Port 003: Dev 002, If 0, Class=Hub, Driver=hub/4p, 5000M'
"""

RE_ID2 = re.compile(r"^(?P<indent2> +)ID (?P<vendor_product>\w\w\w\w:\w\w\w\w)")
"""
Example: '    ID 1d6b:0003 Linux Foundation 3.0 root hub'
"""


def _parse(lsusb_output: str) -> Topology:
    def iter_parse() -> Iterator[Path]:
        lines = lsusb_output.strip().splitlines()
        assert len(lines) % 2 == 0
        current_nesting = 1
        current_bus: None | int = None
        current_path: list[int] = []
        for line1, line2 in itertools.batched(lines, n=2):
            match2 = RE_ID2.match(line2)
            assert match2 is not None
            indent2_len = len(match2.group("indent2"))

            match1 = RE_BUS1.match(line1)
            if match1 is not None:
                # Bus
                assert indent2_len == 4
                current_nesting = 1
                current_bus = int(match1.group("bus"))
                port = int(match1.group("port"))
                assert port == 1
                # print(f"{current_nesting=} {current_bus=} {port=}")
                current_path = []
                continue

            match1 = RE_PORT1.match(line1)
            assert match1 is not None
            # Port
            indent1_len = len(match1.group("indent1"))
            port = int(match1.group("port"))
            device_class = match1.group("class")
            indent2_len = len(match2.group("indent2"))
            assert indent1_len == indent2_len

            new_nesting = indent2_len // 4
            # The nesting must never increase more than one
            assert new_nesting <= new_nesting + 1
            current_nesting = new_nesting

            if device_class == "Hub":
                vendor_product = match2.group("vendor_product")
                current_path = current_path[: current_nesting - 2]
                current_path.append(port)
                # print(
                #     f"{current_nesting=} {current_bus=} {current_path=} {port=} {device_class=} {vendor_product}"
                # )
                yield Path(
                    bus=current_bus,
                    path=current_path,
                    product_id=ProductId.parse(vendor_product),
                )

    return Topology(list(iter_parse()))


def get_real_topology(lsusb_output: str) -> Topology:
    return _parse(lsusb_output=lsusb_output)



from .usbhubctl import BackendPowerABC, Path
from .util_subprocess import subprocess_run


class BackendPowerSysFs(BackendPowerABC):
    def power(self, full_paths: list["Path"], on: bool) -> None:
        """
        /usr/sbin/usbhubctl_sysfs /sys/bus/usb/devices/4-1:1.0/4-1-port1/disable 1
        """
        assert isinstance(full_paths, list)
        for full_path in full_paths:
            assert isinstance(full_path, Path)

        value = "0" if on else "1"

        args = ["usbhubctl_sysfs"]
        for full_path in full_paths:
            # path = f"/sys/bus/usb/devices/{full_path.uhubctl_location}:1.0/{full_path.uhubctl_location}-port{full_path.uhubctl_port}/disable"
            args.append(full_path.sysfs_path)
            args.append(value)
        subprocess_run(args=args)



"""
Sequence:
  Create definition of hub
  Query usb and create hub structure
  Find hub definition in hub structure: Create 'hub mount point'.
  Access hub using 'hub mount point'
"""

import abc
import dataclasses
import logging
import re
from collections.abc import Iterator
from typing import Optional, Union

logger = logging.getLogger(__file__)

# from usbhubctl.backend_power_uhubctl import BackendPowerUhubctl
# backend_power = BackendPowerUhubctl()
# from usbhubctl.backend_power_sysfs import BackendPowerSysFs
# backend_power = BackendPowerSysFs()


class BackendPower:
    _backend_power: Optional["BackendPowerABC"] = None

    @property
    def backend_power(self) -> "BackendPowerABC":
        if self._backend_power is None:
            from usbhubctl.backend_power_sysfs import BackendPowerSysFs

            self._backend_power = BackendPowerSysFs()
        return self._backend_power


_BACKEND_POWER = BackendPower()


class Cache:
    _cache: Optional["Cache"] = None

    def __init__(self):
        self._actual_usb_topology = None

    @property
    def actual_usb_topology(self) -> "Topology":
        if self._actual_usb_topology is None:
            from usbhubctl import backend_query_lsusb

            self._actual_usb_topology = backend_query_lsusb.lsusb()
        return self._actual_usb_topology


_CACHE = Cache()

_RE_PARSE_SHORT_WITH_PRODUCT = re.compile(
    r"^(?P<vendor_product>\w\w\w\w:\w\w\w\w) (?P<bus>\d+)-(?P<path>[\d.]+)$"
)
"""
Example: '05E3:0626 1-3.4.5'
"""


class BackendPowerABC(abc.ABC):
    @abc.abstractmethod
    def power(self, full_paths: list["Path"], on: bool) -> None: ...


@dataclasses.dataclass(frozen=True, repr=True, eq=True)
class ProductId:
    vendor: int
    product: int

    def __post_init__(self) -> None:
        assert isinstance(self.vendor, int)
        assert isinstance(self.product, int)

    @property
    def text(self) -> str:
        return f"{self.vendor:04X}:{self.product:04X}"

    @staticmethod
    def parse(product: str) -> "ProductId":
        """
        Example 'product': "0bda:5411"
        """
        v, p = product.split(":", 2)
        return ProductId(vendor=int(v, base=16), product=int(p, base=16))


@dataclasses.dataclass(frozen=True, repr=True, eq=True)
class DualProductId:
    usb2: ProductId
    usb3: None | ProductId

    def __post_init__(self) -> None:
        assert isinstance(self.usb2, ProductId)
        assert isinstance(self.usb3, None | ProductId)

    def get_product_id(self, is_usb2: bool) -> ProductId:
        return self.usb2 if is_usb2 else self.usb3

    @staticmethod
    def parse(dual_product: str) -> "DualProductId":
        """
        Example 'dual_product': ""0bda:0411-0bda:5411"
        """
        _usb2, _usb3 = dual_product.split("-", 2)
        usb3 = None if (_usb3 == "") else ProductId.parse(_usb3)
        return DualProductId(
            usb2=ProductId.parse(_usb2),
            usb3=usb3,
        )


@dataclasses.dataclass
class Path:
    """
    A USB hub chip and the downstream path leading to this chip.
    """

    product_id: ProductId
    bus: None | int
    path: list[int]

    def __post_init__(self) -> None:
        assert isinstance(self.product_id, ProductId)
        assert isinstance(self.bus, None | int)
        if isinstance(self.path, tuple):
            self.path = list(self.path)
        assert isinstance(self.path, list)

    @property
    def text(self) -> str:
        return f"{','.join(map(str,self.path))}-{self.product_id.text}"

    def is_top_path(self, sub_path: "Path") -> bool:
        assert isinstance(sub_path, Path)
        assert len(sub_path.path) == 0
        return self.product_id == sub_path.product_id

    def equals(self, solution_path: "Path", path: "Path") -> bool:
        assert isinstance(solution_path, Path)
        assert isinstance(path, Path)
        if self.product_id != path.product_id:
            return False
        full_path = solution_path.path + path.path
        return self.path == full_path

    @property
    def short(self) -> str:
        """
        Example:
            4-1
            4-1.4
            4-1.3
            4-1.3
            4-1.3.4
        """
        path = ".".join(map(str, self.path))
        return f"{self.bus}-{path}"

    @property
    def short_with_product(self) -> str:
        return f"{self.product_id.text} {self.short}"

    @staticmethod
    def short_with_product_factory(line: str) -> "Path":
        """
        Example line: '05E3:0626 1-3'
        """
        match = _RE_PARSE_SHORT_WITH_PRODUCT.match(line)
        assert match
        vendor_product = match.group("vendor_product")
        bus = int(match.group("bus"))
        path_text = match.group("path")
        path = [int(port) for port in path_text.split(".")]
        return Path(ProductId.parse(vendor_product), bus, path)

    @property
    def uhubctl_location(self) -> str:
        """
        Example:
            4-1
            4-1.4
            4-1.3
            4-1.3
            4-1.3.4
        """
        path = ".".join(map(str, self.path[:-1]))
        return f"{self.bus}-{path}"

    @property
    def uhubctl_port(self) -> str:
        return f"{self.path[-1]}"

    @property
    def sysfs_path(self) -> str:
        return f"/sys/bus/usb/devices/{self.uhubctl_location}:1.0/{self.uhubctl_location}-port{self.uhubctl_port}/disable"


@dataclasses.dataclass
class ConnectedPlug:
    connected_hub: "ConnectedHub"
    plug: "Plug"

    def power(self, on: bool, backend_power: None | BackendPowerABC = None) -> None:
        # logger.info(f"plug={self.plug.number} power on={on} {full_path}")
        backend_power = _BACKEND_POWER.singleton()
        backend_power.power(full_paths=[self.full_path], on=on)

    @property
    def full_path(self) -> Path:
        root_path = self.connected_hub.root_path
        return Path(
            product_id=root_path.product_id,
            bus=root_path.bus,
            path=root_path.path + self.plug.internal_path,
        )


@dataclasses.dataclass
class DualConnectedPlug:
    connected_plug_usb2: "ConnectedPlug"
    connected_plug_usb3: "ConnectedPlug"

    def __post_init__(self) -> None:
        assert isinstance(self.connected_plug_usb2, ConnectedPlug)
        assert isinstance(self.connected_plug_usb3, ConnectedPlug)

    def power(self, on: bool, backend_power: None | BackendPowerABC = None) -> None:
        # self.connected_plug_usb2.power(on=on, backend_power=backend_power)
        # self.connected_plug_usb3.power(on=on, backend_power=backend_power)

        backend_power = _BACKEND_POWER.backend_power
        full_paths = [
            self.connected_plug_usb2.full_path,
            self.connected_plug_usb3.full_path,
        ]
        backend_power.power(full_paths=full_paths, on=on)


@dataclasses.dataclass
class ConnectedHub:
    hub: "Hub"
    root_path: Path

    @property
    def connected_plugs(self) -> list[ConnectedPlug]:
        return [ConnectedPlug(connected_hub=self, plug=plug) for plug in self.hub.plugs]

    def get_plug(self, plug_number: int) -> ConnectedPlug:
        return ConnectedPlug(connected_hub=self, plug=self.hub.get_plug(plug_number))


@dataclasses.dataclass
class ConnectedHubs:
    hub: "Hub"
    hubs: list[ConnectedHub]

    def __post_init__(self) -> None:
        assert isinstance(self.hub, Hub)
        assert isinstance(self.hubs, list)

    def assert_one(self) -> None:
        if len(self.hubs) > 1:
            print(f"More than one '{self.hub.model}' hub detected: unambiguously")
            return
        if len(self.hubs) == 0:
            print("No hub '{self.hub.model}' detected")
            return

    def get_one(self) -> ConnectedHub:
        self.assert_one()
        return self.hubs[0]

    @property
    def short(self) -> str:
        return "\n".join([hub.root_path.short for hub in self.hubs])


@dataclasses.dataclass
class DualConnectedHub:
    hub: "Hub"
    connected_hub_usb2: ConnectedHub
    connected_hub_usb3: ConnectedHub

    def __post_init__(self) -> None:
        assert isinstance(self.connected_hub_usb2, ConnectedHub)
        assert isinstance(self.connected_hub_usb3, ConnectedHub)

    @property
    def connected_plugs(self) -> list[DualConnectedPlug]:
        return [self.get_plug(plug_number=plug.number) for plug in self.hub.plugs]

    def get_plug(self, plug_number: int) -> DualConnectedPlug:
        return DualConnectedPlug(
            connected_plug_usb2=self.connected_hub_usb2.get_plug(
                plug_number=plug_number
            ),
            connected_plug_usb3=self.connected_hub_usb3.get_plug(
                plug_number=plug_number
            ),
        )


@dataclasses.dataclass
class DualConnectedHubs:
    hub: "Hub"
    hubs_usb2: ConnectedHubs
    hubs_usb3: ConnectedHubs

    def __post_init__(self) -> None:
        assert isinstance(self.hub, Hub)
        assert isinstance(self.hubs_usb2, ConnectedHubs)
        assert isinstance(self.hubs_usb3, ConnectedHubs)

    def assert_one(self) -> None:
        self.hubs_usb2.assert_one()
        self.hubs_usb3.assert_one()

    def get_one(self) -> DualConnectedHub:
        return DualConnectedHub(
            hub=self.hub,
            connected_hub_usb2=self.hubs_usb2.get_one(),
            connected_hub_usb3=self.hubs_usb3.get_one(),
        )

    @property
    def short(self) -> str:
        return "\n".join([hub.root_path.short for hub in self.hubs])

    def assert_one(self) -> None:
        self.hubs_usb2.assert_one()
        self.hubs_usb3.assert_one()


@dataclasses.dataclass
class Topology:
    list_path: list[Path]

    def __post_init__(self) -> None:
        assert isinstance(self.list_path, list)
        self.list_path.sort(key=lambda path: (path.bus, path.path))

    def find_connected_hubs(self, actual_usb_topology: "Topology") -> list[Path]:
        """
        'self' is the hub topology.
        'actual_usb_topology' is the usb topology connected to the PC.
        """

        def try_solution(i_candidate: int) -> None | Path:
            for i, path in enumerate(self.list_path):
                try:
                    candiate = actual_usb_topology.list_path[i_candidate + i]
                except IndexError:
                    return None
                if i == 0:
                    if not candiate.is_top_path(path):
                        return None
                    solution_path = candiate
                    continue

                if not candiate.equals(solution_path, path):
                    return None

            return solution_path

        def iter_solution() -> Iterator[Path]:
            for i_candidate in range(len(actual_usb_topology.list_path)):
                solution_path = try_solution(i_candidate=i_candidate)
                if solution_path is not None:
                    yield solution_path
                    continue

        return list(iter_solution())

    @staticmethod
    def short_with_product_factory(multiline: str) -> "Topology":
        multiline = multiline.strip()
        list_path = [
            Path.short_with_product_factory(line) for line in multiline.splitlines()
        ]
        return Topology(list_path=list_path)

    @property
    def short(self) -> str:
        """
        Example:
            4-1
            4-1.4
            4-1.3
            4-1.3
            4-1.3.4
        """
        return "\n".join([path.short for path in self.list_path])

    @property
    def short_with_product(self) -> str:
        """
        Example:
            4-1
            4-1.4
            4-1.3
            4-1.3
            4-1.3.4
        """
        return "\n".join([path.short_with_product for path in self.list_path])


@dataclasses.dataclass
class HubChip:
    """
    A USB Hub chip.
    In the 'USB Spec', this would be a 'USB controller'.
    There is exactly one upstream port and typically 4 downstream ports.
    Every downstram port is connected to a downstream usb hub chip or to
    a plug on the 'Hub'.
    """

    product_id: None | DualProductId
    plug_or_chip: list[Union[int, "HubChip"]]

    def __post_init__(self) -> None:
        assert isinstance(self.product_id, None | DualProductId)
        assert isinstance(self.plug_or_chip, list)
        for port in self.plug_or_chip:
            assert isinstance(port, int | HubChip)

    def _register(self, hub: "Hub", path: list[int]) -> None:
        for port_number0, plug_or_chip in enumerate(self.plug_or_chip):
            _path = path + [port_number0 + 1]
            if isinstance(plug_or_chip, int):
                hub._register_plug(Plug(number=plug_or_chip, internal_path=_path))
                continue

            assert isinstance(plug_or_chip, HubChip)
            plug_or_chip._register(hub, _path)

    def get_product_id(self, is_usb2: bool) -> ProductId:
        return self.product_id.get_product_id(is_usb2=is_usb2)


@dataclasses.dataclass
class Plug:
    number: int
    internal_path: list[int]

    def __post_init__(self) -> None:
        assert isinstance(self.number, int)
        if isinstance(self.internal_path, tuple):
            self.internal_path = list(self.internal_path)
        assert isinstance(self.internal_path, list)


@dataclasses.dataclass
class Hub:
    """
    A Hub consists of one more more 'Chips'.
    """

    manufacturer: str
    model: str
    comment: str
    plug_count: int
    hub_chip: HubChip

    _dict_plugs: dict[int, Plug] = dataclasses.field(default_factory=dict)
    """
    The key is the plug number (starting from 1).
    """

    def __post_init__(self) -> None:
        assert isinstance(self.manufacturer, str)
        assert isinstance(self.model, str)
        assert isinstance(self.plug_count, int)
        assert isinstance(self.hub_chip, HubChip)

        self.hub_chip._register(self, path=[])

        sorted_ports = sorted(self._dict_plugs.keys())
        if len(sorted_ports) != self.plug_count:
            raise ValueError(
                f"Expected {self.plug_count} plugs, but only {len(sorted_ports)} plugs have been defined: {sorted_ports}"
            )

        for port0, port in enumerate(sorted_ports):
            if port0 + 1 != port:
                raise ValueError(
                    f"Not all ports from 1 to {self.plug_count} have been defined: {sorted_ports}"
                )

    @property
    def plugs(self) -> list[Plug]:
        return [
            self._dict_plugs[plug_number]
            for plug_number in sorted(self._dict_plugs.keys())
        ]

    def _register_plug(self, plug: Plug) -> None:
        assert plug.number not in self._dict_plugs
        self._dict_plugs[plug.number] = plug

    def get_plug(self, plug_number: int) -> Plug:
        return self._dict_plugs[plug_number]

    def get_topology(self, is_usb2: bool) -> Topology:
        list_internal_path: list[Path] = []

        def downstream(hub_chip: "HubChip", path: list[int]) -> None:
            assert hub_chip.product_id is not None

            list_internal_path.append(
                Path(
                    product_id=hub_chip.get_product_id(is_usb2=is_usb2),
                    bus=None,
                    path=path,
                )
            )
            for port_number0, plug_or_chip in enumerate(hub_chip.plug_or_chip):
                if isinstance(plug_or_chip, HubChip):
                    _path = path + [port_number0 + 1]
                    downstream(plug_or_chip, _path)

        downstream(self.hub_chip, path=[])

        return Topology(list_internal_path)

    def find_connected_hubs(
        self,
        is_usb2: bool,
        actual_usb_topology: Optional["Topology"] = None,
    ) -> ConnectedHubs:
        assert isinstance(actual_usb_topology, None | Topology)
        assert isinstance(is_usb2, bool)

        if actual_usb_topology is None:
            actual_usb_topology = _CACHE.actual_usb_topology

        return ConnectedHubs(
            hub=self,
            hubs=[
                ConnectedHub(hub=self, root_path=path)
                for path in self.get_topology(is_usb2=is_usb2).find_connected_hubs(
                    actual_usb_topology=actual_usb_topology
                )
            ],
        )

    def find_connected_dualhubs(
        self,
        actual_usb_topology: Optional["Topology"] = None,
    ) -> DualConnectedHubs:
        return DualConnectedHubs(
            hub=self,
            hubs_usb2=self.find_connected_hubs(
                is_usb2=True, actual_usb_topology=actual_usb_topology
            ),
            hubs_usb3=self.find_connected_hubs(
                is_usb2=False, actual_usb_topology=actual_usb_topology
            ),
        )

    def rotate_toggle(self, root_path: Path) -> None:
        for plug_number in sorted(self._dict_plugs.keys()):
            plug = self._dict_plugs[plug_number]
            full_path = plug.extend(root_path=root_path)
            print(full_path)

    def __repr__(self) -> str:
        return f"USB Hub '{self.model}' with {self.plug_count} plugs"



from usbhubctl import Path
from usbhubctl.usbhubctl import BackendPowerABC

from .util_subprocess import subprocess_run


class BackendPowerUhubctl(BackendPowerABC):
    def power(self, full_paths: list["Path"], on: bool) -> None:
        """
        uhubctl -l 3-6.1 -p 1 --action on
        """
        # TODO: This is too much simplified!
        assert len(full_paths) == 1
        full_path = full_paths[0]

        action = "on" if on else "off"
        args = [
            "uhubctl",
            f"--location={full_path.uhubctl_location}",
            f"--ports={full_path.uhubctl_port}",
            # f"--vendor={full_path.product_id.text}",
            # "--nosysfs",
            f"--action={action}",
        ]
        subprocess_run(args=args)



"""
https://en.wikipedia.org/wiki/USB
https://github.com/pyusb/pyusb/blob/master/docs/tutorial.rst
"""

import dataclasses
import time
from typing import Dict, List
import usb.core
import usb.util
from usb.legacy import CLASS_HUB


"""
Hub(vendor=8523, product=29264, bus=3, path=(5, 2, 1), address=9)
Hub(vendor=8457, product=2066, bus=3, path=(5, 2), address=6)
Hub(vendor=6720, product=2049, bus=3, path=(5,), address=4)

lsusb -tv
/:  Bus 003.Port 001: Dev 001, Class=root_hub, Driver=xhci_hcd/12p, 480M
    ID 1d6b:0002 Linux Foundation 2.0 root hub
    |__ Port 005: Dev 004, If 0, Class=Hub, Driver=hub/4p, 480M
        ID 1a40:0801 Terminus Technology Inc. 
        |__ Port 002: Dev 006, If 0, Class=Hub, Driver=hub/4p, 480M
            ID 2109:0812 VIA Labs, Inc. VL812 Hub
            |__ Port 001: Dev 009, If 0, Class=Hub, Driver=hub/4p, 480M
                ID 214b:7250 Huasheng Electronics 

ls -1 /sys/bus/usb/devices
3-5
3-5:1.0
3-5.2
3-5.2.1
3-5.2:1.0
3-5.2.1.1
3-5.2.1:1.0
3-5.2.1.1:1.0
3-5.2.1.4
3-5.2.1.4:1.0

Example 3-5.2.1

readlink -f /sys/bus/usb/devices/3-5.2.1
/sys/devices/pci0000:00/0000:00:14.0/usb3/3-5/3-5.2/3-5.2.1

cat /sys/devices/pci0000:00/0000:00:14.0/usb3/3-5/3-5.2/3-5.2.1/idVendor
214b

cat /sys/devices/pci0000:00/0000:00:14.0/usb3/3-5/3-5.2/3-5.2.1/bDeviceClass
09     # Hub
"""


class find_class:
    def __init__(self, class_: int):
        self._class = class_

    def __call__(self, device):
        # first, let's check the device
        if device.bDeviceClass == self._class:
            return True
        # ok, transverse all devices to find an
        # interface that matches our class
        for cfg in device:
            # find_descriptor: what's it?
            intf = usb.util.find_descriptor(cfg, bInterfaceClass=self._class)
            if intf is not None:
                return True

        return False


@dataclasses.dataclass
class Hub:
    vendor: int
    product: int
    bus: int
    path: List[int]
    address: int


def main():
    begin_s = time.monotonic()

    hub_devices = usb.core.find(find_all=1, custom_match=find_class(CLASS_HUB))
    hub_devices = list(hub_devices)

    hubs: List[Hub] = []
    for hub in hub_devices:
        # print(f"  port_number={hub.port_number}({hub.port_numbers})")
        # print(hub)
        if hub.port_numbers is None:
            continue
        h = Hub(
            vendor=hub.idVendor,
            product=hub.idProduct,
            bus=hub.bus,
            path=hub.port_numbers,
            address=hub.address,
        )
        hubs.append(h)

    for h in hubs:
        print(h)
    print(f"*********** duration={time.monotonic()-begin_s:0.3f}s")


if __name__ == "__main__":
    main()



"""
Specify the topology of a Hub
"""

from usbhubctl import DualProductId, Hub, HubChip

_PRODUCT = DualProductId.parse("2109:0822-2109:2822")

rsh_a107 = Hub(
    manufacturer="RSHTECH",
    model="RSH-A107",
    comment="Power control does NOT work!",
    plug_count=7,
    hub_chip=HubChip(
        _PRODUCT,
        plug_or_chip=[
            HubChip(_PRODUCT, plug_or_chip=[7, 6, 5, 4]),
            3,
            2,
            1,
        ],
    ),
)



__version__ = "0.0.1"

__all__ = [
    "Hub",
    "HubChip",
    "Path",
    "ProductId",
    "DualProductId",
    "DualConnectedHub",
    "DualConnectedHubs",
    "Topology",
    "ConnectedHub",
    "ConnectedHubs",
    "ConnectedPlug",
    "BackendPowerABC",
]

from .usbhubctl import (
    Hub,
    HubChip,
    Path,
    ProductId,
    DualConnectedHub,
    DualConnectedHubs,
    DualProductId,
    Topology,
    ConnectedHub,
    ConnectedHubs,
    ConnectedPlug,
    BackendPowerABC,
)
from .parser_lsusb import get_real_topology



import json
import logging
import logging.config
import pathlib

DIRECTORY_OF_THIS_FILE = pathlib.Path(__file__).parent
FILENAME_LOGGING_JSON = DIRECTORY_OF_THIS_FILE / "util_logging.json"


def init_logging() -> None:
    logging.config.dictConfig(json.loads(FILENAME_LOGGING_JSON.read_text()))



"""
Specify the topology of a Hub
"""

from usbhubctl import DualProductId, Hub, HubChip

_PRODUCT = DualProductId.parse("0bda:0411-0bda:5411")

rsh_a16 = Hub(
    manufacturer="RSHTECH",
    model="RSH-A16",
    comment="",
    plug_count=16,
    hub_chip=HubChip(
        _PRODUCT,
        plug_or_chip=[
            1,
            2,
            HubChip(
                _PRODUCT,
                plug_or_chip=[
                    7,
                    8,
                    HubChip(_PRODUCT, plug_or_chip=[13, 14, 15, 16]),
                    HubChip(_PRODUCT, plug_or_chip=[9, 10, 11, 12]),
                ],
            ),
            HubChip(_PRODUCT, plug_or_chip=[3, 4, 5, 6]),
        ],
    ),
)



"""
https://en.wikipedia.org/wiki/USB
https://github.com/pyusb/pyusb/blob/master/docs/tutorial.rst
"""

import dataclasses
import time

import usb.core
import usb.util
import usb.control
import usb.legacy


class FindUsbClass:
    def __init__(self, class_: int):
        self._class = class_

    def __call__(self, device: usb.core.Device) -> bool:
        # first, let's check the device
        if device.bDeviceClass == self._class:
            return True
        # ok, transverse all devices to find an
        # interface that matches our class
        for cfg in device:
            # find_descriptor: what's it?
            intf = usb.util.find_descriptor(cfg, bInterfaceClass=self._class)
            if intf is not None:
                return True

        return False


@dataclasses.dataclass
class Hub:
    vendor: int
    product: int
    bus: int
    path: list[int]
    address: int

    @property
    def pattern(self) -> tuple[list[int], str]:
        return (self.path, f"{self.vendor:04X}:{self.product:04X}")


def main() -> None:
    begin_s = time.monotonic()

    hub_devices = usb.core.find(
        find_all=1, custom_match=FindUsbClass(usb.legacy.CLASS_HUB)
    )
    hub_devices = list(hub_devices)

    dev = usb.core.find(idVendor=0x0BDA, idProduct=0x0411)
    print("_str(): ", dev._str())
    print("_get_full_descriptor_str(): ", dev._get_full_descriptor_str())
    print("dev", dev)

    def get_descriptor(dev, desc_size, desc_type, desc_index, wIndex=0):
        r"""Return the specified descriptor.

        dev is the Device object to which the request will be
        sent to.

        desc_size is the descriptor size.

        desc_type and desc_index are the descriptor type and index,
        respectively. wIndex index is used for string descriptors
        and represents the Language ID. For other types of descriptors,
        it is zero.
        """
        wValue = desc_index | (desc_type << 8)

        bmRequestType = usb.util.build_request_type(
            usb.util.CTRL_IN,
            usb.util.CTRL_TYPE_STANDARD,
            usb.util.CTRL_RECIPIENT_DEVICE,
        )

        desc = dev.ctrl_transfer(
            bmRequestType=bmRequestType,
            bRequest=0x06,
            wValue=wValue,
            wIndex=wIndex,
            data_or_wLength=desc_size,
        )

        if len(desc) < 2:
            raise usb.core.USBError("Invalid descriptor")

        return desc

    desc = get_descriptor(
        dev, desc_size=2, desc_type=usb.legacy.DT_HUB << 8, desc_index=0
    )
    print("desc:", desc)
    return

    # ret = dev.ctrl_transfer(0xC0, CTRL_LOOPBACK_READ, 0, 0, len(msg))

    # dev.set_configuration()

    cfg: usb.core.Configuration = dev.get_active_configuration()
    print("cfg: ", cfg)
    print("type(cfg): ", type(cfg))
    intf: usb.core.Interface = cfg[(0, 0)]

    print("intf: ", intf)
    print("type(intf): ", type(intf))

    return

    hubs: list[Hub] = []
    for hub in hub_devices:
        # print(f"  port_number={hub.port_number}({hub.port_numbers})")
        # print(hub)
        if hub.port_numbers is None:
            continue

        if (hub.idVendor == 0x0BDA) and (hub.idProduct == 0x0411):
            cfg = hub.get_active_configuration()
            intf = cfg[(0, 0)]

            alt = usb.util.find_descriptor(cfg, find_all=True, bInterfaceNumber=0)
            assert alt is not None
            for a in alt:
                print(f"{a=} {type(a)}")
                for i in a:
                    print(f"{i=}")

            if False:
                ep = usb.util.find_descriptor(
                    intf,
                    # match the first OUT endpoint
                    custom_match=lambda e: usb.util.endpoint_direction(
                        e.bEndpointAddress
                    )
                    == usb.util.ENDPOINT_OUT,
                )

                assert ep is not None

                # # write the data
                # ep.write("test")

        h = Hub(
            vendor=hub.idVendor,
            product=hub.idProduct,
            bus=hub.bus,
            path=hub.port_numbers,
            address=hub.address,
        )
        hubs.append(h)

    list_pattern = sorted([h.pattern for h in hubs])
    for pattern in list_pattern:
        print(pattern)
    print(f"*********** duration={time.monotonic()-begin_s:0.3f}s")


if __name__ == "__main__":
    main()



from usbhubctl import Topology, get_real_topology

LSUSB_OUTPUT_RSH_A16 = """
/:  Bus 001.Port 001: Dev 001, Class=root_hub, Driver=xhci_hcd/1p, 480M
    ID 1d6b:0002 Linux Foundation 2.0 root hub
/:  Bus 002.Port 001: Dev 001, Class=root_hub, Driver=xhci_hcd/4p, 10000M
    ID 1d6b:0003 Linux Foundation 3.0 root hub
    |__ Port 003: Dev 002, If 0, Class=Hub, Driver=hub/4p, 5000M
        ID 05e3:0626 Genesys Logic, Inc. Hub
        |__ Port 001: Dev 003, If 0, Class=Hub, Driver=hub/4p, 5000M
            ID 0bda:0411 Realtek Semiconductor Corp. Hub
            |__ Port 003: Dev 005, If 0, Class=Hub, Driver=hub/4p, 5000M
                ID 0bda:0411 Realtek Semiconductor Corp. Hub
                |__ Port 003: Dev 007, If 0, Class=Hub, Driver=hub/4p, 5000M
                    ID 0bda:0411 Realtek Semiconductor Corp. Hub
                |__ Port 004: Dev 008, If 0, Class=Hub, Driver=hub/4p, 5000M
                    ID 0bda:0411 Realtek Semiconductor Corp. Hub
            |__ Port 004: Dev 006, If 0, Class=Hub, Driver=hub/4p, 5000M
                ID 0bda:0411 Realtek Semiconductor Corp. Hub
        |__ Port 004: Dev 004, If 0, Class=Vendor Specific Class, Driver=r8152, 5000M
            ID 0bda:8153 Realtek Semiconductor Corp. RTL8153 Gigabit Ethernet Adapter
/:  Bus 003.Port 001: Dev 001, Class=root_hub, Driver=xhci_hcd/12p, 480M
    ID 1d6b:0002 Linux Foundation 2.0 root hub
    |__ Port 001: Dev 002, If 0, Class=Human Interface Device, Driver=usbhid, 12M
        ID 046d:c534 Logitech, Inc. Unifying Receiver
    |__ Port 001: Dev 002, If 1, Class=Human Interface Device, Driver=usbhid, 12M
        ID 046d:c534 Logitech, Inc. Unifying Receiver
    |__ Port 003: Dev 003, If 0, Class=Vendor Specific Class, Driver=[none], 12M
        ID 06cb:00f9 Synaptics, Inc. 
    |__ Port 004: Dev 004, If 0, Class=Video, Driver=uvcvideo, 480M
        ID 13d3:5405 IMC Networks 
    |__ Port 004: Dev 004, If 1, Class=Video, Driver=uvcvideo, 480M
        ID 13d3:5405 IMC Networks 
    |__ Port 004: Dev 004, If 2, Class=Video, Driver=uvcvideo, 480M
        ID 13d3:5405 IMC Networks 
    |__ Port 004: Dev 004, If 3, Class=Video, Driver=uvcvideo, 480M
        ID 13d3:5405 IMC Networks 
    |__ Port 004: Dev 004, If 4, Class=Application Specific Interface, Driver=[none], 480M
        ID 13d3:5405 IMC Networks 
    |__ Port 006: Dev 005, If 0, Class=Hub, Driver=hub/4p, 480M
        ID 05e3:0610 Genesys Logic, Inc. Hub
        |__ Port 001: Dev 007, If 0, Class=Hub, Driver=hub/4p, 480M
            ID 0bda:5411 Realtek Semiconductor Corp. RTS5411 Hub
            |__ Port 003: Dev 008, If 0, Class=Hub, Driver=hub/4p, 480M
                ID 0bda:5411 Realtek Semiconductor Corp. RTS5411 Hub
                |__ Port 003: Dev 010, If 0, Class=Hub, Driver=hub/4p, 480M
                    ID 0bda:5411 Realtek Semiconductor Corp. RTS5411 Hub
                |__ Port 004: Dev 011, If 0, Class=Hub, Driver=hub/4p, 480M
                    ID 0bda:5411 Realtek Semiconductor Corp. RTS5411 Hub
            |__ Port 004: Dev 009, If 0, Class=Hub, Driver=hub/4p, 480M
                ID 0bda:5411 Realtek Semiconductor Corp. RTS5411 Hub
/:  Bus 004.Port 001: Dev 001, Class=root_hub, Driver=xhci_hcd/4p, 10000M
    ID 1d6b:0003 Linux Foundation 3.0 root hub
"""

LSUSB_OUTPUT_RSH_A16_EXPECTED_SHORT = """
2-3
2-3.1
2-3.1.3
2-3.1.3.3
2-3.1.3.4
2-3.1.4
3-6
3-6.1
3-6.1.3
3-6.1.3.3
3-6.1.3.4
3-6.1.4
""".strip()


LSUSB_OUTPUT_RSH_A10 = """
/:  Bus 001.Port 001: Dev 001, Class=root_hub, Driver=xhci_hcd/1p, 480M
    ID 1d6b:0002 Linux Foundation 2.0 root hub
/:  Bus 002.Port 001: Dev 001, Class=root_hub, Driver=xhci_hcd/4p, 10000M
    ID 1d6b:0003 Linux Foundation 3.0 root hub
    |__ Port 003: Dev 002, If 0, Class=Hub, Driver=hub/4p, 5000M
        ID 05e3:0626 Genesys Logic, Inc. Hub
        |__ Port 003: Dev 003, If 0, Class=Hub, Driver=hub/4p, 5000M
            ID 0bda:0411 Realtek Semiconductor Corp. Hub
            |__ Port 003: Dev 005, If 0, Class=Hub, Driver=hub/4p, 5000M
                ID 0bda:0411 Realtek Semiconductor Corp. Hub
            |__ Port 004: Dev 006, If 0, Class=Hub, Driver=hub/4p, 5000M
                ID 0bda:0411 Realtek Semiconductor Corp. Hub
        |__ Port 004: Dev 004, If 0, Class=Vendor Specific Class, Driver=r8152, 5000M
            ID 0bda:8153 Realtek Semiconductor Corp. RTL8153 Gigabit Ethernet Adapter
/:  Bus 003.Port 001: Dev 001, Class=root_hub, Driver=xhci_hcd/12p, 480M
    ID 1d6b:0002 Linux Foundation 2.0 root hub
    |__ Port 003: Dev 003, If 0, Class=Vendor Specific Class, Driver=[none], 12M
        ID 06cb:00f9 Synaptics, Inc. 
    |__ Port 004: Dev 004, If 0, Class=Video, Driver=uvcvideo, 480M
        ID 13d3:5405 IMC Networks 
    |__ Port 004: Dev 004, If 1, Class=Video, Driver=uvcvideo, 480M
        ID 13d3:5405 IMC Networks 
    |__ Port 004: Dev 004, If 2, Class=Video, Driver=uvcvideo, 480M
        ID 13d3:5405 IMC Networks 
    |__ Port 004: Dev 004, If 3, Class=Video, Driver=uvcvideo, 480M
        ID 13d3:5405 IMC Networks 
    |__ Port 004: Dev 004, If 4, Class=Application Specific Interface, Driver=[none], 480M
        ID 13d3:5405 IMC Networks 
    |__ Port 006: Dev 006, If 0, Class=Hub, Driver=hub/4p, 480M
        ID 05e3:0610 Genesys Logic, Inc. Hub
        |__ Port 003: Dev 007, If 0, Class=Hub, Driver=hub/4p, 480M
            ID 0bda:5411 Realtek Semiconductor Corp. RTS5411 Hub
            |__ Port 001: Dev 008, If 0, Class=Mass Storage, Driver=usb-storage, 12M
                ID 2e8a:0003  
            |__ Port 001: Dev 008, If 1, Class=Vendor Specific Class, Driver=[none], 12M
                ID 2e8a:0003  
            |__ Port 003: Dev 009, If 0, Class=Hub, Driver=hub/4p, 480M
                ID 0bda:5411 Realtek Semiconductor Corp. RTS5411 Hub
            |__ Port 004: Dev 010, If 0, Class=Hub, Driver=hub/4p, 480M
                ID 0bda:5411 Realtek Semiconductor Corp. RTS5411 Hub
    |__ Port 007: Dev 012, If 0, Class=Human Interface Device, Driver=usbhid, 12M
        ID 046d:c534 Logitech, Inc. Unifying Receiver
    |__ Port 007: Dev 012, If 1, Class=Human Interface Device, Driver=usbhid, 12M
        ID 046d:c534 Logitech, Inc. Unifying Receiver
/:  Bus 004.Port 001: Dev 001, Class=root_hub, Driver=xhci_hcd/4p, 10000M
    ID 1d6b:0003 Linux Foundation 3.0 root hub
"""

LSUSB_OUTPUT_RSH_A10_EXPECTED_SHORT = """
2-3
2-3.3
2-3.3.3
2-3.3.4
3-6
3-6.3
3-6.3.3
3-6.3.4
""".strip()


def test_parser_lsusb_A16():
    topology = get_real_topology(lsusb_output=LSUSB_OUTPUT_RSH_A16)
    assert topology.short == LSUSB_OUTPUT_RSH_A16_EXPECTED_SHORT


def test_parser_lsusb_A10():
    topology = get_real_topology(lsusb_output=LSUSB_OUTPUT_RSH_A10)
    assert topology.short == LSUSB_OUTPUT_RSH_A10_EXPECTED_SHORT



"""
Specify the topology of a Hub
"""

from usbhubctl import DualProductId, Hub, HubChip

_PRODUCT = DualProductId.parse("0bda:0411-0bda:5411")

rsh_a10 = Hub(
    manufacturer="RSHTECH",
    model="RSH-A10",
    comment="",
    plug_count=10,
    hub_chip=HubChip(
        _PRODUCT,
        plug_or_chip=[
            1,
            2,
            HubChip(_PRODUCT, plug_or_chip=[7, 8, 9, 10]),
            HubChip(_PRODUCT, plug_or_chip=[3, 4, 5, 6]),
        ],
    ),
)
